import logging
from .context import context_register


class Context2Record(logging.Filter):

    def __init__(self, *a, ctx_name, **kw):
        self.ctx_name = ctx_name
        super().__init__(*a, **kw)

    def filter(self, record):
        ctx_cls = context_register.get(self.ctx_name)
        for name, value in ctx_cls.dict().items():
            setattr(record, name, value)
        return True



import logging.config
import uuid
import asyncio
from aiohttp import web

from asyncio_context import Context


class ViewsCtx(Context):
    user_id: str = 'UNKNOWN'
    trace_id: str = 'NO_TRACE'


ctx = ViewsCtx()
log = logging.getLogger()


async def another_task():
    log.info('another task')


class MyView(web.View):

    async def get(self):
        ViewsCtx().user_id = self.request.match_info.get('user_id')
        ctx.trace_id = self.request.headers.get('X-Trace-Id', str(uuid.uuid4()))
        log.info('Test request')

        task = asyncio.get_running_loop().create_task(another_task())
        ViewsCtx.copy_to_task(task)

        # To show same trace id in aiohttp exception log
        raise Exception()

        return web.Response(body='hello world!')


app = web.Application()
app.router.add_route("GET", "/user/{user_id}", MyView)


LOGGING_CONFIG = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'standard': {
            'format': '[trace_id:%(trace_id)s] [user_id:%(user_id)s]:%(message)s'
        },
    },
    'filters': {
        'context2record': {
            '()': 'asyncio_context.logging.Context2Record',
            'ctx_name': 'ViewsCtx'
        },
    },
    'handlers': {
        'default': {
            'level': 'INFO',
            'formatter': 'standard',
            'class': 'logging.StreamHandler',
            'filters': ['context2record'],
        },
    },
    'loggers': {
        '': {
            'handlers': ['default'],
            'level': 'DEBUG',
            'propagate': False,
        },
    }
}
logging.config.dictConfig(LOGGING_CONFIG)
log.info('Starting...')
web.run_app(app)



import asyncio
from typing import Dict


TASK_CONTEXT_ATTR = 'asyncio_context'
context_register: Dict[str, 'Context'] = {}


def _get_raw_ctx() -> dict:
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        return {}
    task = asyncio.current_task(loop=loop)
    return getattr(task, TASK_CONTEXT_ATTR, {})


class _Attrs2Context(type):

    @staticmethod
    def make_setter(name):
        def setter(self, value):
            task = asyncio.current_task()
            ctx_name = f'{self._prefix}{name}'
            if not hasattr(task, TASK_CONTEXT_ATTR):
                new_ctx = {
                    f'{self._prefix}{k}': v for k, v in self._origin.items()}
                setattr(task, TASK_CONTEXT_ATTR, new_ctx)
            getattr(task, TASK_CONTEXT_ATTR)[ctx_name] = value

        return setter

    @staticmethod
    def make_getter(name):
        def getter(self):
            ctx = _get_raw_ctx()
            ctx_name = f'{self._prefix}{name}'
            if ctx_name not in ctx:
                return self._origin[name]
            return ctx[ctx_name]

        return getter

    def __new__(mcs, name, bases, attrs):
        new_attrs = {'_origin': {}}
        for attr_name, attr_value in attrs.items():
            if attr_name.startswith('_') or attr_name in (
                    'dict', 'copy_to_task'):
                new_attrs[attr_name] = attr_value
                continue
            new_attrs['_origin'][attr_name] = attr_value
            new_attrs[attr_name] = property(
                fget=mcs.make_getter(attr_name),
                fset=mcs.make_setter(attr_name))
        return super().__new__(mcs, name, bases, new_attrs)


class Context(metaclass=_Attrs2Context):
    _origin: dict
    _prefix: str = ''

    def __init_subclass__(cls, **kwargs):
        context_register[cls.__qualname__] = cls

    @classmethod
    def dict(cls) -> dict:
        ctx = _get_raw_ctx()
        res = {}
        for k, v in cls._origin.items():
            ctx_name = f'{cls._prefix}{k}'
            res[k] = ctx.get(ctx_name, cls._origin[k])
        return res

    @classmethod
    def copy_to_task(cls, task: asyncio.Task):
        setattr(task, TASK_CONTEXT_ATTR, cls.dict())
